from .constants import SchemaURI


class SCIMException(Exception):
    status = 500
    schema = SchemaURI.ERROR
    scim_type = None

    def __init__(self, detail=None, **kwargs):
        self.detail = detail or ''
        self.status = kwargs.get('status') or self.status
        self.schemas = kwargs.get('schemas') or [self.schema]
        self.scim_type = kwargs.get('scim_type') or self.scim_type

        msg = '({} {}) {}'.format(self.status, self.scim_type, self.detail)

        super(Exception, self).__init__(msg)

    def to_dict(self):
        d = {
            'schemas': self.schemas,
            'detail': self.detail,
            'status': self.status,
        }
        if self.scim_type:
            d['scimType'] = self.scim_type

        return d


class AuthorizationError(SCIMException):
    status = 401


class NotFoundError(SCIMException):
    status = 404

    def __init__(self, uuid, **kwargs):
        detail = u'Resource {} not found'.format(uuid)
        super(NotFoundError, self).__init__(detail, **kwargs)


class BadRequestError(SCIMException):
    status = 400


class IntegrityError(SCIMException):
    status = 409


class NotImplementedError(SCIMException):
    status = 501



from urllib.parse import urljoin

from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from . import constants
from . import exceptions
from .settings import scim_settings
from .utils import get_base_scim_location_getter


class SCIMServiceProviderConfig(object):
    """
    A reference ServiceProviderConfig. This should be overridden to
    describe those authentication_schemes and features that are implemented by
    your app.
    """
    def __init__(self, request=None):
        self.request = request

    @property
    def meta(self):
        return {
            'location': self.location,
            'resourceType': 'ServiceProviderConfig',
        }

    @property
    def location(self):
        path = reverse('scim:service-provider-config')
        return urljoin(get_base_scim_location_getter()(self.request), path)

    def to_dict(self):
        return {
            'schemas': [constants.SchemaURI.SERVICE_PROVIDER_CONFIG],
            'documentationUri': scim_settings.DOCUMENTATION_URI,
            'patch': {
                'supported': True,
            },
            'bulk': {
                'supported': False,
                'maxOperations': 1000,
                'maxPayloadSize': 1048576,
            },
            # Django-SCIM2 does not fully support the SCIM2.0 filtering spec.
            # Until it does, let's under promise and over deliver to the world.
            'filter': {
                'supported': False,
                'maxResults': 50,
            },
            'changePassword': {
                'supported': True,
            },
            'sort': {
                'supported': False,
            },
            'etag': {
                'supported': False,
            },
            'authenticationSchemes': scim_settings.AUTHENTICATION_SCHEMES,
            'meta': self.meta,
        }


class AbstractSCIMCommonAttributesMixin(models.Model):
    """
    An abstract model to provide SCIM Common Attributes.

    https://tools.ietf.org/html/rfc7643#section-3.1

    Each SCIM resource (Users, Groups, etc.) includes the following
    common attributes.  With the exception of the "ServiceProviderConfig"
    and "ResourceType" server discovery endpoints and their associated
    resources, these attributes MUST be defined for all resources,
    including any extended resource types.  When accepted by a service
    provider (e.g., after a SCIM create), the attributes "id" and "meta"
    (and its associated sub-attributes) MUST be assigned values by the
    service provider.  Common attributes are considered to be part of
    every base resource schema and do not use their own "schemas" URI.

    For backward compatibility, some existing schema definitions MAY list
    common attributes as part of the schema.  The attribute
    characteristics (see Section 2.2) listed here SHALL take precedence
    over older definitions that may be included in existing schemas.
    """

    """
    id
      A unique identifier for a SCIM resource as defined by the service
      provider.  Each representation of the resource MUST include a
      non-empty "id" value.  This identifier MUST be unique across the
      SCIM service provider's entire set of resources.  It MUST be a
      stable, non-reassignable identifier that does not change when the
      same resource is returned in subsequent requests.  The value of
      the "id" attribute is always issued by the service provider and
      MUST NOT be specified by the client.  The string "bulkId" is a
      reserved keyword and MUST NOT be used within any unique identifier
      value.  The attribute characteristics are "caseExact" as "true", a
      mutability of "readOnly", and a "returned" characteristic of
      "always".  See Section 9 for additional considerations regarding
      privacy.
    """
    scim_id = models.CharField(
        _('SCIM ID'),
        max_length=254,
        null=True,
        blank=True,
        default=None,
        db_index=True,
        help_text=_('A unique identifier for a SCIM resource as defined by the service provider.'),
    )

    """
    externalId
      A String that is an identifier for the resource as defined by the
      provisioning client.  The "externalId" may simplify identification
      of a resource between the provisioning client and the service
      provider by allowing the client to use a filter to locate the
      resource with an identifier from the provisioning domain,
      obviating the need to store a local mapping between the
      provisioning domain's identifier of the resource and the
      identifier used by the service provider.  Each resource MAY
      include a non-empty "externalId" value.  The value of the
      "externalId" attribute is always issued by the provisioning client
      and MUST NOT be specified by the service provider.  The service
      provider MUST always interpret the externalId as scoped to the
      provisioning domain.  While the server does not enforce
      uniqueness, it is assumed that the value's uniqueness is
      controlled by the client setting the value.  See Section 9 for
      additional considerations regarding privacy.  This attribute has
      "caseExact" as "true" and a mutability of "readWrite".  This
      attribute is OPTIONAL.
    """
    scim_external_id = models.CharField(
        _('SCIM External ID'),
        max_length=254,
        null=True,
        blank=True,
        default=None,
        db_index=True,
        help_text=_('A string that is an identifier for the resource as defined by the provisioning client.'),
    )

    def set_scim_id(self, is_new):
        if is_new:
            self.__class__.objects.filter(id=self.id).update(scim_id=self.id)
            self.scim_id = str(self.id)

    def save(self, *args, **kwargs):
        is_new = self.id is None
        super(AbstractSCIMCommonAttributesMixin, self).save(*args, **kwargs)
        self.set_scim_id(is_new)

    class Meta:
        abstract = True


class AbstractSCIMUserMixin(AbstractSCIMCommonAttributesMixin):
    """
    An abstract model to provide the User resource schema.

    # https://tools.ietf.org/html/rfc7643#section-4.1
    """

    """
    userName
      A service provider's unique identifier for the user, typically
      used by the user to directly authenticate to the service provider.
      Often displayed to the user as their unique identifier within the
      system (as opposed to "id" or "externalId", which are generally
      opaque and not user-friendly identifiers).  Each User MUST include
      a non-empty userName value.  This identifier MUST be unique across
      the service provider's entire set of Users.  This attribute is
      REQUIRED and is case insensitive.
    """
    scim_username = models.CharField(
        _('SCIM Username'),
        max_length=254,
        null=True,
        blank=True,
        default=None,
        db_index=True,
        help_text=_("A service provider's unique identifier for the user"),
    )

    @property
    def scim_groups(self):
        raise exceptions.NotImplementedError

    class Meta:
        abstract = True


class AbstractSCIMGroupMixin(AbstractSCIMCommonAttributesMixin):
    """
    An abstract model to provide the Group resource schema.

    # https://tools.ietf.org/html/rfc7643#section-4.2
    """

    """
    displayName
      A human-readable name for the Group.  REQUIRED.
    """
    scim_display_name = models.CharField(
        _('SCIM Display Name'),
        max_length=254,
        null=True,
        blank=True,
        default=None,
        db_index=True,
        help_text=_("A human-readable name for the Group."),
    )

    class Meta:
        abstract = True

    def set_scim_display_name(self, is_new):
        if is_new:
            self.__class__.objects.filter(id=self.id).update(scim_display_name=self.name)
            self.scim_display_name = self.name

    def save(self, *args, **kwargs):
        is_new = self.id is None
        super(AbstractSCIMGroupMixin, self).save(*args, **kwargs)
        self.set_scim_display_name(is_new)




import json
import logging
from urllib.parse import urljoin

from django.contrib.auth import get_user_model
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django import db
from django.db import transaction
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from django.utils.decorators import method_decorator
from django.urls import reverse

from . import constants
from . import exceptions
from .utils import get_all_schemas_getter
from .utils import get_group_adapter
from .utils import get_group_model
from .utils import get_user_adapter
from .utils import get_user_filter_parser
from .utils import get_group_filter_parser
from .utils import get_base_scim_location_getter
from .utils import get_service_provider_config_model
from .utils import get_extra_model_filter_kwargs_getter
from .utils import get_extra_model_exclude_kwargs_getter


logger = logging.getLogger(__name__)


class SCIMView(View):
    lookup_url_kwarg = 'uuid'  # argument in django URL pattern
    implemented = True

    @property
    def lookup_field(self):
        """Database field, possibly redefined in the adapter"""
        return getattr(self.scim_adapter, 'id_field', 'scim_id')

    @property
    def model_cls(self):
        # pull from __class__ to avoid binding model class getter to
        # self instance and passing self to class getter
        return self.__class__.model_cls_getter()

    @property
    def get_extra_filter_kwargs(self):
        return get_extra_model_filter_kwargs_getter(self.model_cls)

    @property
    def get_extra_exclude_kwargs(self):
        return get_extra_model_exclude_kwargs_getter(self.model_cls)

    @property
    def scim_adapter(self):
        # pull from __class__ to avoid binding adapter class getter to
        # self instance and passing self to class getter
        return self.__class__.scim_adapter_getter()

    def get_object(self):
        """Get object by configurable ID."""
        # Perform the lookup filtering.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        assert lookup_url_kwarg in self.kwargs, (
            'Expected view %s to be called with a URL keyword argument '
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            'attribute on the view correctly.' %
            (self.__class__.__name__, lookup_url_kwarg)
        )

        uuid = self.kwargs[lookup_url_kwarg]

        extra_filter_kwargs = self.get_extra_filter_kwargs(self.request, uuid)
        extra_filter_kwargs[self.lookup_field] = uuid
        # No use of get_extra_exclude_kwargs here since we are
        # searching for a specific single object.

        try:
            return self.model_cls.objects.get(**extra_filter_kwargs)
        except ObjectDoesNotExist:
            raise exceptions.NotFoundError(uuid)

    @method_decorator(csrf_exempt)
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        if not self.implemented:
            return self.status_501(request, *args, **kwargs)

        try:
            return super(SCIMView, self).dispatch(request, *args, **kwargs)
        except Exception as e:
            if not isinstance(e, exceptions.SCIMException):
                logger.exception('Unable to complete SCIM call.')
                e = exceptions.SCIMException(str(e))

            content = json.dumps(e.to_dict())
            return HttpResponse(content=content,
                                content_type=constants.SCIM_CONTENT_TYPE,
                                status=e.status)

    def status_501(self, request, *args, **kwargs):
        """
        A service provider that does NOT support a feature SHOULD
        respond with HTTP status code 501 (Not Implemented).
        """
        return HttpResponse(content_type=constants.SCIM_CONTENT_TYPE, status=501)

    def load_body(self, body):
        decoded = body.decode(constants.ENCODING)
        stripped = decoded.strip() or '{}'

        try:
            return json.loads(stripped)
        except json.decoder.JSONDecodeError as e:
            msg = 'Could not decode JSON body: ' + e.args[0]
            raise exceptions.BadRequestError(msg)


class FilterMixin(object):

    parser_getter = None
    scim_adapter_getter = None

    def _page(self, request):
        try:
            start = request.GET.get('startIndex', 1)
            if start is not None:
                start = int(start)
                if start < 1:
                    raise exceptions.BadRequestError('Invalid startIndex (must be >= 1)')

            count = request.GET.get('count', 50)
            if count is not None:
                count = int(count)

            return start, count

        except ValueError as e:
            raise exceptions.BadRequestError('Invalid pagination values: ' + str(e))

    def _search(self, request, query, start, count):
        try:
            qs = self.__class__.parser_getter().search(query, request)
        except ValueError as e:
            raise exceptions.BadRequestError('Invalid filter/search query: ' + str(e))

        extra_filter_kwargs = self.get_extra_filter_kwargs(request)
        qs = self._filter_raw_queryset_with_extra_filter_kwargs(qs, extra_filter_kwargs)
        extra_exclude_kwargs = self.get_extra_exclude_kwargs(request)
        qs = self._filter_raw_queryset_with_extra_exclude_kwargs(qs, extra_exclude_kwargs)

        return self._build_response(request, qs, start, count)

    def _filter_raw_queryset_with_extra_filter_kwargs(self, qs, extra_filter_kwargs):

        obj_list = []
        for obj in qs:
            add_obj = True
            for attr_name, attr_val in extra_filter_kwargs.items():
                if attr_name.endswith('__in'):
                    attr_name = attr_name.replace('__in', '')
                    if not hasattr(obj, attr_name) or getattr(obj, attr_name) not in attr_val:
                        add_obj = False
                        break

                else:
                    if not hasattr(obj, attr_name) or getattr(obj, attr_name) != attr_val:
                        add_obj = False
                        break

            if add_obj:
                obj_list.append(obj)

        return obj_list

    def _filter_raw_queryset_with_extra_exclude_kwargs(self, qs, extra_exclude_kwargs):

        obj_list = []
        for obj in qs:
            add_obj = True
            for attr_name, attr_val in extra_exclude_kwargs.items():
                if attr_name.endswith('__in'):
                    attr_name = attr_name.replace('__in', '')
                    if hasattr(obj, attr_name) and getattr(obj, attr_name) in attr_val:
                        add_obj = False
                        break

                else:
                    if hasattr(obj, attr_name) and getattr(obj, attr_name) == attr_val:
                        add_obj = False
                        break

            if add_obj:
                obj_list.append(obj)

        return obj_list

    def _build_response(self, request, qs, start, count):
        try:
            total_count = sum(1 for _ in qs)
            qs = qs[start - 1:(start - 1) + count]
            resources = [self.scim_adapter(o, request=request).to_dict() for o in qs]
            doc = {
                'schemas': [constants.SchemaURI.LIST_RESPONSE],
                'totalResults': total_count,
                'itemsPerPage': count,
                'startIndex': start,
                'Resources': resources,
            }
        except ValueError as e:
            raise exceptions.BadRequestError(str(e))
        else:
            content = json.dumps(doc)
            return HttpResponse(content=content,
                                content_type=constants.SCIM_CONTENT_TYPE)


class SearchView(FilterMixin, SCIMView):
    http_method_names = ['post']

    # override model class so correct extra_filter/exclude_kwarg getter is fetched
    model_cls = 'search'

    def post(self, request):
        body = self.load_body(request.body)
        if body.get('schemas') != [constants.SchemaURI.SERACH_REQUEST]:
            raise exceptions.BadRequestError('Invalid schema uri. Must be SearchRequest.')

        query = body.get('filter', request.GET.get('filter'))

        if not query:
            raise exceptions.BadRequestError('No filter query specified')

        response = self._search(request, query, *self._page(request))
        path = reverse(self.scim_adapter.url_name)
        url = urljoin(get_base_scim_location_getter()(request=request), path).rstrip('/')
        response['Location'] = url + '/.search'
        return response


class UserSearchView(SearchView):
    scim_adapter_getter = get_user_adapter
    parser_getter = get_user_filter_parser


class GroupSearchView(SearchView):
    scim_adapter_getter = get_group_adapter
    parser_getter = get_group_filter_parser


class GetView(object):
    def get(self, request, *args, **kwargs):
        if kwargs.get(self.lookup_url_kwarg):
            return self.get_single(request)

        return self.get_many(request)

    def get_single(self, request):
        obj = self.get_object()
        scim_obj = self.scim_adapter(obj, request=request)
        content = json.dumps(scim_obj.to_dict())
        response = HttpResponse(content=content,
                                content_type=constants.SCIM_CONTENT_TYPE)
        response['Location'] = scim_obj.location
        return response

    def get_many(self, request):
        query = request.GET.get('filter')
        if query:
            return self._search(request, query, *self._page(request))

        extra_filter_kwargs = self.get_extra_filter_kwargs(request)
        extra_exclude_kwargs = self.get_extra_exclude_kwargs(request)
        qs = self.model_cls.objects.filter(
            **extra_filter_kwargs
        ).exclude(
            **extra_exclude_kwargs
        ).order_by(self.lookup_field)
        return self._build_response(request, qs, *self._page(request))


class DeleteView(object):
    def delete(self, request, *args, **kwargs):
        obj = self.get_object()

        scim_obj = self.scim_adapter(obj, request=request)

        scim_obj.delete()

        return HttpResponse(status=204)


class PostView(object):
    def post(self, request, **kwargs):
        obj = self.model_cls()
        scim_obj = self.scim_adapter(obj, request=request)

        body = self.load_body(request.body)

        if not body:
            raise exceptions.BadRequestError('POST call made with empty body')

        scim_obj.from_dict(body)

        try:
            scim_obj.save()
        except db.utils.IntegrityError as e:
            # Cast error to a SCIM IntegrityError to use the status
            # attribute on the SCIM IntegrityError.
            raise exceptions.IntegrityError(str(e))

        content = json.dumps(scim_obj.to_dict())
        response = HttpResponse(content=content,
                                content_type=constants.SCIM_CONTENT_TYPE,
                                status=201)
        response['Location'] = scim_obj.location
        return response


class PutView(object):
    def put(self, request, *args, **kwargs):
        obj = self.get_object()

        scim_obj = self.scim_adapter(obj, request=request)

        body = self.load_body(request.body)

        if not body:
            raise exceptions.BadRequestError('PUT call made with empty body')

        scim_obj.from_dict(body)
        scim_obj.save()

        content = json.dumps(scim_obj.to_dict())
        response = HttpResponse(content=content,
                                content_type=constants.SCIM_CONTENT_TYPE)
        response['Location'] = scim_obj.location
        return response


class PatchView(object):
    def patch(self, request, *args, **kwargs):
        obj = self.get_object()

        scim_obj = self.scim_adapter(obj, request=request)
        body = self.load_body(request.body)

        operations = body.get('Operations')

        if not operations:
            raise exceptions.BadRequestError('PATCH call made without operations array')

        with transaction.atomic():
            scim_obj.handle_operations(operations)

        content = json.dumps(scim_obj.to_dict())
        response = HttpResponse(content=content,
                                content_type=constants.SCIM_CONTENT_TYPE)
        response['Location'] = scim_obj.location
        return response


class UsersView(FilterMixin, GetView, PostView, PutView, PatchView, DeleteView, SCIMView):

    http_method_names = ['get', 'post', 'put', 'patch', 'delete']

    scim_adapter_getter = get_user_adapter
    model_cls_getter = get_user_model
    parser_getter = get_user_filter_parser


class GroupsView(FilterMixin, GetView, PostView, PutView, PatchView, DeleteView, SCIMView):

    http_method_names = ['get', 'post', 'put', 'patch', 'delete']

    scim_adapter_getter = get_group_adapter
    model_cls_getter = get_group_model
    parser_getter = get_group_filter_parser


class ServiceProviderConfigView(SCIMView):
    http_method_names = ['get']

    def get(self, request):
        config = get_service_provider_config_model()(request=request)
        content = json.dumps(config.to_dict())
        return HttpResponse(content=content,
                            content_type=constants.SCIM_CONTENT_TYPE)


class ResourceTypesView(SCIMView):

    http_method_names = ['get']

    def type_dict_by_type_id(self, request):
        type_adapters = get_user_adapter(), get_group_adapter()
        type_dicts = [m.resource_type_dict(request) for m in type_adapters]
        return {d['id']: d for d in type_dicts}

    def get(self, request, uuid=None, *args, **kwargs):
        if uuid:
            doc = self.type_dict_by_type_id(request).get(uuid)
            if not doc:
                return HttpResponse(content_type=constants.SCIM_CONTENT_TYPE, status=404)

        else:
            key_func = lambda o: o.get('id')  # noqa: E731
            type_dicts = self.type_dict_by_type_id(request).values()
            types = list(sorted(type_dicts, key=key_func))
            doc = {
                'schemas': [constants.SchemaURI.LIST_RESPONSE],
                'Resources': types,
            }

        return HttpResponse(content=json.dumps(doc),
                            content_type=constants.SCIM_CONTENT_TYPE)


class SchemasView(SCIMView):

    http_method_names = ['get']

    schemas_by_uri = {s['id']: s for s in get_all_schemas_getter()()}

    def get(self, request, uuid=None, *args, **kwargs):
        if uuid:
            doc = self.schemas_by_uri.get(uuid)
            if not doc:
                return HttpResponse(content_type=constants.SCIM_CONTENT_TYPE, status=404)

        else:
            key_func = lambda o: o.get('id')  # noqa: E731
            schemas = list(sorted(self.schemas_by_uri.values(), key=key_func))
            doc = {
                'schemas': [constants.SchemaURI.LIST_RESPONSE],
                'Resources': schemas,
            }

        content = json.dumps(doc)
        return HttpResponse(content=content,
                            content_type=constants.SCIM_CONTENT_TYPE)




import logging

from django.http.response import HttpResponse
from django.urls import reverse

from . import constants
from .settings import scim_settings
from .utils import get_loggable_body


logger = logging.getLogger(__name__)


class SCIMAuthCheckMiddleware(object):
    """
    Check to see if a prior middleware has logged the user in.

    This middleware should be place after auth middleware used to login a user.
    """

    def __init__(self, get_response=None):
        # One-time configuration and initialization per server start.
        self.get_response = get_response

    def __call__(self, request):
        response = None
        if hasattr(self, 'process_request'):
            response = self.process_request(request)
        if not response:
            response = self.get_response(request)
        if hasattr(self, 'process_response'):
            response = self.process_response(request, response)
        return response

    @property
    def reverse_url(self):
        if not hasattr(self, '_reverse_url'):
            self._reverse_url = reverse('scim:root')
        return self._reverse_url

    def process_request(self, request):
        if request.path.startswith(self.reverse_url):
            self.log_request(request)
        # If we've just passed through the auth middleware and there is no user
        # associated with the request we can assume permission
        # was denied and return a 401.
        if not hasattr(request, 'user') or request.user.is_anonymous:
            if request.path.startswith(self.reverse_url):
                response = HttpResponse(status=401)
                response['WWW-Authenticate'] = scim_settings.WWW_AUTHENTICATE_HEADER
                return response

    def process_response(self, request, response):
        if request.path.startswith(self.reverse_url):
            self.log_response(request, response)
        return response

    def get_loggable_content(self, content):
        try:
            body = get_loggable_body(content.decode(constants.ENCODING))
        except Exception as e:
            body = 'Could not parse request body\n' + str(e)

        return body

    def get_loggable_request_message(self, request):
        body = self.get_loggable_content(request.body)
        parts = [
            'PATH',
            request.path,
            'METHOD',
            request.method,
            'BODY',
            body,
        ]

        return '\n'.join(parts)

    def get_loggable_response_message(self, request, response):
        body = self.get_loggable_content(response.content)
        parts = [
            'PATH',
            request.path,
            'METHOD',
            request.method,
            'BODY',
            body,
            'STATUS_CODE',
            str(response.status_code),
        ]

        return '\n'.join(parts)

    def log_request(self, request):
        message = self.get_loggable_request_message(request)
        logger.debug(message)

    def log_response(self, request, response):
        message = self.get_loggable_response_message(request, response)
        logger.debug(message)



"""
Transform filter query into QuerySet
"""
from django.contrib.auth import get_user_model

from scim2_filter_parser.queries.sql import SQLQuery

from .utils import get_group_model


class FilterQuery:
    model_getter = None
    joins = ()
    attr_map = None
    query_class = SQLQuery

    @classmethod
    def table_name(cls):
        return cls.model_getter()._meta.db_table

    @classmethod
    def search(cls, filter_query, request=None):
        q = cls.query_class(filter_query, cls.table_name(), cls.attr_map, cls.joins)
        if q.where_sql is None:
            return cls.model_getter().objects.none()

        sql, params = cls.get_raw_args(q, request)

        return cls.model_getter().objects.raw(sql, params)

    @classmethod
    def get_raw_args(cls, q, request=None):
        """
        Return a Query object's SQL augmented with params from cls.get_extras.
        """
        sql, params = q.sql, q.params

        extra_sql, extra_params = cls.get_extras(q, request)
        if extra_sql:
            if "'%s'" in extra_sql:
                raise ValueError(
                    'Dangerous use of quotes around place holder. Please see '
                    'https://docs.djangoproject.com/en/2.2/ref/models/querysets/#extra '
                    'for more details.'
                )

            sql = sql.rstrip(';') + extra_sql + ';'
            params += extra_params

        return sql, params

    @classmethod
    def get_extras(cls, q, request=None) -> (str, list):
        """
        Return extra SQL and params to be attached to end of current Query's
        SQL and params.

        For example:
            return 'AND tenant_id = %s', [request.user.tenant_id]
        """
        return '', []


class UserFilterQuery(FilterQuery):
    model_getter = get_user_model
    attr_map = {}


class GroupFilterQuery(FilterQuery):
    model_getter = get_group_model
    attr_map = {}




import json
from urllib.parse import urlunparse

from .settings import scim_settings


def get_user_adapter():
    """
    Return the user model adapter.
    """
    return scim_settings.USER_ADAPTER


def get_group_model():
    """
    Return the group model.
    """
    return scim_settings.GROUP_MODEL


def get_group_adapter():
    """
    Return the group model adapter.
    """
    return scim_settings.GROUP_ADAPTER


def get_user_filter_parser():
    """
    Return the user filter parser.
    """
    return scim_settings.USER_FILTER_PARSER


def get_group_filter_parser():
    """
    Return the group filter parser.
    """
    return scim_settings.GROUP_FILTER_PARSER


def get_service_provider_config_model():
    """
    Return the Service Provider Config model.
    """
    return scim_settings.SERVICE_PROVIDER_CONFIG_MODEL


def get_base_scim_location_getter():
    """
    Return a function that will, when called, returns the base
    location of scim app.
    """
    return scim_settings.BASE_LOCATION_GETTER


def get_all_schemas_getter():
    """
    Return a function that will, when called, returns the base
    location of scim app.
    """
    return scim_settings.SCHEMAS_GETTER


def get_extra_model_filter_kwargs_getter(model):
    """
    Return a function that will, when called, returns the base
    location of scim app.
    """
    return scim_settings.GET_EXTRA_MODEL_FILTER_KWARGS_GETTER(model)


def get_extra_model_exclude_kwargs_getter(model):
    """
    Return a function that will, when called, returns the base
    location of scim app.
    """
    return scim_settings.GET_EXTRA_MODEL_EXCLUDE_KWARGS_GETTER(model)


def default_base_scim_location_getter(request=None, *args, **kwargs):
    """
    Return the default location of the app implementing the SCIM api.
    """
    base_scim_location_parts = (
        scim_settings.SCHEME,
        scim_settings.NETLOC,
        '',  # path
        '',  # params
        '',  # query
        ''   # fragment
    )

    base_scim_location = urlunparse(base_scim_location_parts)

    return base_scim_location


def default_get_extra_model_filter_kwargs_getter(model):
    """
    Return a **method** that will return extra model filter kwargs for the passed in model.

    :param model:
    """
    def get_extra_filter_kwargs(request, *args, **kwargs):
        """
        Return extra filter kwargs for the given model.
        :param request:
        :param args:
        :param kwargs:
        :rtype: dict
        """
        return {}

    return get_extra_filter_kwargs


def default_get_extra_model_exclude_kwargs_getter(model):
    """
    Return a **method** that will return extra model exclude kwargs for the passed in model.

    :param model:
    """
    def get_extra_exclude_kwargs(request, *args, **kwargs):
        """
        Return extra exclude kwargs for the given model.
        :param request:
        :param args:
        :param kwargs:
        :rtype: dict
        """
        return {}

    return get_extra_exclude_kwargs


def clean_structure_of_passwords(obj):
    if isinstance(obj, dict):
        new_obj = {}
        for key, value in obj.items():
            if 'password' in key.lower():
                new_obj[key] = '*' * len(value) if value else None
            else:
                new_obj[key] = clean_structure_of_passwords(value)

        return new_obj

    elif isinstance(obj, list):
        return [clean_structure_of_passwords(item) for item in obj]

    else:
        return obj


def get_loggable_body(text):
    if not text:
        return text

    try:
        obj = json.loads(text)
    except json.JSONDecodeError:
        return text

    obj = clean_structure_of_passwords(obj)

    return json.dumps(obj)




import re


ENCODING = 'utf-8'
SCIM_CONTENT_TYPE = 'application/scim+json'
VALID_PATCH_OPS = ('add', 'remove', 'replace')


class SchemaURI(object):
    ERROR = 'urn:ietf:params:scim:api:messages:2.0:Error'
    LIST_RESPONSE = 'urn:ietf:params:scim:api:messages:2.0:ListResponse'
    SERACH_REQUEST = 'urn:ietf:params:scim:api:messages:2.0:SearchRequest'
    NOT_SERACH_REQUEST = 'urn:ietf:params:scim:api:messages:2.0:NotSearchRequest'
    PATCH_OP = 'urn:ietf:params:scim:api:messages:2.0:PatchOp'

    USER = 'urn:ietf:params:scim:schemas:core:2.0:User'
    ENTERPRISE_USER = 'urn:ietf:params:scim:schemas:extension:enterprise:2.0:User'
    GROUP = 'urn:ietf:params:scim:schemas:core:2.0:Group'
    RESOURCE_TYPE = 'urn:ietf:params:scim:schemas:core:2.0:ResourceType'
    SERVICE_PROVIDER_CONFIG = 'urn:ietf:params:scim:schemas:core:2.0:ServiceProviderConfig'




"""
Adapters are used to convert the data model described by the SCIM 2.0
specification to a data model that fits the data provided by the application
implementing a SCIM api.

For example, in a Django app, there are User and Group models that do
not have the same attributes/fields that are defined by the SCIM 2.0
specification. The Django User model has both ``first_name`` and ``last_name``
attributes but the SCIM speicifcation requires this same data be sent under
the names ``givenName`` and ``familyName`` respectively.

An adapter is instantiated with a model instance. Eg::

    user = get_user_model().objects.get(id=1)
    scim_user = SCIMUser(user)
    ...

"""
from typing import Optional, Union
from urllib.parse import urljoin

from django.contrib.auth import get_user_model
from django.urls import reverse
from django import core
from scim2_filter_parser.attr_paths import AttrPaths

from . import constants
from . import exceptions
from . import types
from .utils import get_base_scim_location_getter
from .utils import get_group_adapter
from .utils import get_user_adapter


class SCIMMixin(object):

    ATTR_MAP = {}

    id_field = 'scim_id'  # Modifiable by overriding classes

    def __init__(self, obj, request=None):
        self.obj = obj
        self._request = request

    @property
    def request(self):
        if self._request:
            return self._request

        raise RuntimeError('Adapter is not associated with a request object. '
                           'Set object.request to avoid this error.')

    @request.setter
    def request(self, value):
        self._request = value

    @property
    def id(self):
        return str(getattr(self.obj, self.id_field))

    @property
    def path(self):
        return reverse(self.url_name, kwargs={'uuid': self.id})

    @property
    def location(self):
        return urljoin(get_base_scim_location_getter()(self.request), self.path)

    def to_dict(self):
        """
        Return a ``dict`` conforming to the object's SCIM Schema,
        ready for conversion to a JSON object.
        """
        d = {
            'id': self.id,
            'externalId': self.obj.scim_external_id,
        }

        return d

    def from_dict(self, d):
        """
        Consume a ``dict`` conforming to the object's SCIM Schema, updating the
        internal object with data from the ``dict``.

        This method is overridden and called by subclass adapters. Please make
        changes there.
        """
        scim_external_id = d.get('externalId')
        self.obj.scim_external_id = scim_external_id or ''

    def save(self):
        self.obj.save()

    def delete(self):
        self.obj.__class__.objects.filter(id=self.id).delete()

    def handle_operations(self, operations):
        """
        The SCIM specification allows for making changes to specific attributes
        of a model. These changes are sent in PATCH requests and are batched into
        operations to be performed on a object. Operations can have an op code
        of 'add', 'remove', or 'replace'. This method iterates through all of the
        operations in ``operations`` and calls the appropriate handler (defined
        on the appropriate adapter) for each.

        Django-scim2 only provides a partial implementation of PATCH call
        handlers. The RFC (https://tools.ietf.org/html/rfc7644#section-3.5.2)
        specifies a number of requirements for a full PATCH implementation.
        This implementation does not meet all of those requirements. For
        example, these are some features that have been left out.

        Add Operations:
            - If the target location does not exist, the attribute and value
              are added.
        Remove Operations:
            - If the target location is a multi-valued attribute and a complex
              filter is specified comparing a "value", the values matched by the
              filter are removed.  If no other values remain after removal of
              the selected values, the multi-valued attribute SHALL be
              considered unassigned.
        Replace Operations:
            - If the target location path specifies an attribute that does not
              exist, the service provider SHALL treat the operation as an "add".
        """
        for operation in operations:
            path = operation.get('path')
            value = operation.get('value')

            path, value = self.parse_path_and_value(path, value)

            op_code = operation.get('op').lower()

            if op_code not in constants.VALID_PATCH_OPS:
                raise exceptions.BadRequestError(f'Unknown PATCH op "{op_code}"')

            if op_code == 'remove' and not path:
                msg = f'"path" must be specified during "remove" PATCH calls'
                raise exceptions.BadRequestError(msg, scim_type='noTarget')

            op_code = 'handle_' + op_code
            handler = getattr(self, op_code)

            handler(path, value, operation)

    def parse_path_and_value(self,
                             path: Optional[str],
                             value: Union[str, list, dict]) -> tuple:
        """
        Return new path and value given an original path and value.

        This method can be overridden to provide a more usable path and value
        within the associated handle methods.
        """
        # Convert all path's to 3-tuple (attr, subattr, Uri) in preparation for
        # use of scim2-filter-parser. Complex paths can path through as the logic
        # to handle them is not in place yet.
        if not path and isinstance(value, dict):
            # If there is no path and value is a dict, we assume that each
            # key in the dict is an attribute path. Let's convert attribute
            # paths to 3-tuples to have a uniform API.
            new_path = None
            new_value = {}
            for k, v in value.items():
                new_value[self.split_path(k)] = v
        else:
            new_path = self.split_path(path)
            new_value = value

        return new_path, new_value

    def split_path(self, path: str) -> Union[tuple, types.ComplexAttrPath]:
        """
        Convert path to 3-tuple of (attr, subattr, uri) if possible.
        """
        # AttrPaths requires a complete filter query. Thus we tack on
        # ' eq ""' to path to make a complete SCIM query.
        filter_ = path + ' eq ""'
        attr_paths = list(AttrPaths(filter_, self.ATTR_MAP))

        if len(attr_paths) == 0:
            msg = 'No attribute path found in request'
            raise scim_exceptions.BadRequestError(msg)

        if len(attr_paths) > 1:
            # When in doubt hand it all back.
            return self.handle_complex_attr_path(path, attr_paths)

        return attr_paths[0]

    def handle_complex_attr_path(self, path: str, attr_paths: list) -> types.ComplexAttrPath:
        """
        Return an object that has complex path context.

        It's up to the handlers to reject, ignore, handle requests with
        these types of paths. Handling them is above and beyond what
        the maintainer has time for.

            - "addresses[type eq "work"]"
            - "members[value eq "123"].displayName"
            - "emails[type eq "work" and value co "@example.com"].value"
        """
        return types.ComplexAttrPath(path, attr_paths)

    def handle_add(self,
                   path: Union[tuple, types.ComplexAttrPath],
                   value: Union[str, list, dict],
                   operation: dict):
        """
        Handle add operations per:
        https://tools.ietf.org/html/rfc7644#section-3.5.2.1
        """
        raise exceptions.NotImplementedError

    def handle_remove(self,
                      path: tuple,
                      value: Union[str, list, dict],
                      operation: dict):
        """
        Handle remove operations per:
        https://tools.ietf.org/html/rfc7644#section-3.5.2.2
        """
        raise exceptions.NotImplementedError

    def handle_replace(self,
                       path: Union[tuple, types.ComplexAttrPath],
                       value: Union[str, list, dict],
                       operation: dict):
        """
        Handle replace operations per:
        https://tools.ietf.org/html/rfc7644#section-3.5.2.3
        """
        raise exceptions.NotImplementedError


class SCIMUser(SCIMMixin):
    """
    Adapter for adding SCIM functionality to a Django User object.

    This adapter can be overridden; see the ``USER_ADAPTER`` setting
    for details.
    """
    # not great, could be more decoupled. But \__( )__/ whatevs.
    url_name = 'scim:users'
    resource_type = 'User'

    # In order to start consolidating the frameworks for referencing
    # attributes with scim2-filter-parser, a attribute map like
    # the one used in that library is created here.
    ATTR_MAP = {
        # attr, sub attr, uri
        ('userName', None, None): 'username',
        ('name', 'familyName', None): 'last_name',
        ('familyName', None, None): 'last_name',
        ('name', 'givenName', None): 'first_name',
        ('givenName', None, None): 'first_name',
        ('active', None, None): 'is_active',
    }

    @property
    def display_name(self):
        """
        Return the displayName of the user per the SCIM spec.
        """
        if self.obj.first_name and self.obj.last_name:
            return u'{0.first_name} {0.last_name}'.format(self.obj)
        return self.obj.username

    @property
    def name_formatted(self):
        return self.display_name

    @property
    def emails(self):
        """
        Return the email of the user per the SCIM spec.
        """
        return [{'value': self.obj.email, 'primary': True}]

    @property
    def groups(self):
        """
        Return the groups of the user per the SCIM spec.
        """
        group_qs = self.obj.scim_groups.all()
        scim_groups = [get_group_adapter()(g, self.request) for g in group_qs]

        dicts = []
        for group in scim_groups:
            d = {
                'value': group.id,
                '$ref': group.location,
                'display': group.display_name,
            }
            dicts.append(d)

        return dicts

    @property
    def meta(self):
        """
        Return the meta object of the user per the SCIM spec.
        """
        d = {
            'resourceType': self.resource_type,
            'created': self.obj.date_joined.isoformat(),
            'lastModified': self.obj.date_joined.isoformat(),
            'location': self.location,
        }

        return d

    def to_dict(self):
        """
        Return a ``dict`` conforming to the SCIM User Schema,
        ready for conversion to a JSON object.
        """
        d = super().to_dict()
        d.update({
            'schemas': [constants.SchemaURI.USER],
            'userName': self.obj.username,
            'name': {
                'givenName': self.obj.first_name,
                'familyName': self.obj.last_name,
                'formatted': self.name_formatted,
            },
            'displayName': self.display_name,
            'emails': self.emails,
            'active': self.obj.is_active,
            'groups': self.groups,
            'meta': self.meta,
        })

        return d

    def from_dict(self, d):
        """
        Consume a ``dict`` conforming to the SCIM User Schema, updating the
        internal user object with data from the ``dict``.

        Please note, the user object is not saved within this method. To
        persist the changes made by this method, please call ``.save()`` on the
        adapter. Eg::

            scim_user.from_dict(d)
            scim_user.save()
        """
        super().from_dict(d)

        username = d.get('userName')
        self.obj.username = username or ''

        self.obj.scim_username = self.obj.username

        first_name = d.get('name', {}).get('givenName')
        self.obj.first_name = first_name or ''

        last_name = d.get('name', {}).get('familyName')
        self.obj.last_name = last_name or ''

        emails = d.get('emails', [])
        self.parse_emails(emails)

        cleartext_password = d.get('password')
        if cleartext_password:
            self.obj.set_password(cleartext_password)

        active = d.get('active')
        if active is not None:
            self.obj.is_active = active

    @classmethod
    def resource_type_dict(cls, request=None):
        """
        Return a ``dict`` containing ResourceType metadata for the user object.
        """
        id_ = cls.resource_type
        path = reverse('scim:resource-types', kwargs={'uuid': id_})
        location = urljoin(get_base_scim_location_getter()(request), path)
        return {
            'schemas': [constants.SchemaURI.RESOURCE_TYPE],
            'id': id_,
            'name': 'User',
            'endpoint': reverse('scim:users'),
            'description': 'User Account',
            'schema': constants.SchemaURI.USER,
            'meta': {
                'location': location,
                'resourceType': 'ResourceType'
            }
        }

    def parse_emails(self, value: Optional[list]):
        if value:
            email = None
            if isinstance(value, list):
                primary_emails = sorted(
                    (e for e in value if e.get('primary')),
                    key=lambda d: d.get('value')
                )
                secondary_emails = sorted(
                    (e for e in value if not e.get('primary')),
                    key=lambda d: d.get('value')
                )

                emails = primary_emails + secondary_emails
                if emails:
                    email = emails[0].get('value')
                else:
                    raise exceptions.BadRequestError('Invalid email value')

            elif isinstance(value, dict):
                # if value is a dict, let's assume it contains the primary email.
                # OneLogin sends a dict despite the spec:
                #   https://tools.ietf.org/html/rfc7643#section-4.1.2
                #   https://tools.ietf.org/html/rfc7643#section-8.2
                email = (value.get('value') or '').strip()

            self.validate_email(email)

            self.obj.email = email

    @staticmethod
    def validate_email(email):
        try:
            core.validators.EmailValidator()(email)
        except core.exceptions.ValidationError:
            raise exceptions.BadRequestError('Invalid email value')

    def handle_replace(self,
                       path: Optional[tuple],
                       value: Union[str, list, dict],
                       operation: dict):
        """
        Handle the replace operations.
        """
        if path and isinstance(value, str):
            # Restructure for use in loop below.
            value = {path: value}

        if not isinstance(value, dict):
            raise exceptions.NotImplementedError(
                'PATCH replace operation with value type of '
                '{type(value)} is not implemented'
            )

        for path, value in (value or {}).items():
            if path in self.ATTR_MAP:
                setattr(self.obj, self.ATTR_MAP.get(path), value)

            elif path == ('emails', None, None):
                self.parse_emails(value)

            else:
                raise exceptions.NotImplementedError('Not Implemented')

        self.obj.save()


class SCIMGroup(SCIMMixin):
    """
    Adapter for adding SCIM functionality to a Django Group object.

    This adapter can be overridden; see the ``GROUP_ADAPTER``
    setting for details.
    """
    # not great, could be more decoupled. But \__( )__/ whatevs.
    url_name = 'scim:groups'
    resource_type = 'Group'

    @property
    def display_name(self):
        """
        Return the displayName of the group per the SCIM spec.
        """
        return self.obj.name

    @property
    def members(self):
        """
        Return a list of user dicts (ready for serialization) for the members
        of the group.

        :rtype: list
        """
        users = self.obj.user_set.all()
        scim_users = [get_user_adapter()(user, self.request) for user in users]

        dicts = []
        for user in scim_users:
            d = {
                'value': user.id,
                '$ref': user.location,
                'display': user.display_name,
            }
            dicts.append(d)

        return dicts

    @property
    def meta(self):
        """
        Return the meta object of the group per the SCIM spec.
        """
        d = {
            'resourceType': self.resource_type,
            'location': self.location,
        }

        return d

    def to_dict(self):
        """
        Return a ``dict`` conforming to the SCIM Group Schema,
        ready for conversion to a JSON object.
        """
        d = super().to_dict()
        d.update({
            'schemas': [constants.SchemaURI.GROUP],
            'displayName': self.display_name,
            'members': self.members,
            'meta': self.meta,
        })
        return d

    def from_dict(self, d):
        """
        Consume a ``dict`` conforming to the SCIM Group Schema, updating the
        internal group object with data from the ``dict``.

        Please note, the group object is not saved within this method. To
        persist the changes made by this method, please call ``.save()`` on the
        adapter. Eg::

            scim_group.from_dict(d)
            scim_group.save()
        """
        super().from_dict(d)

        name = d.get('displayName')
        self.obj.name = name or ''

    @classmethod
    def resource_type_dict(cls, request=None):
        """
        Return a ``dict`` containing ResourceType metadata for the group object.
        """
        id_ = cls.resource_type
        path = reverse('scim:resource-types', kwargs={'uuid': id_})
        location = urljoin(get_base_scim_location_getter()(request), path)
        return {
            'schemas': [constants.SchemaURI.RESOURCE_TYPE],
            'id': id_,
            'name': 'Group',
            'endpoint': reverse('scim:groups'),
            'description': 'Group',
            'schema': constants.SchemaURI.GROUP,
            'meta': {
                'location': location,
                'resourceType': 'ResourceType'
            }
        }

    def handle_add(self, path, value, operation):
        """
        Handle add operations.
        """
        if path == ('members', None, None):
            members = value or []
            ids = [int(member.get('value')) for member in members]
            users = get_user_model().objects.filter(id__in=ids)

            if len(ids) != users.count():
                raise exceptions.BadRequestError('Can not add a non-existent user to group')

            for user in users:
                self.obj.user_set.add(user)

        else:
            raise exceptions.NotImplementedError

    def handle_remove(self, path, value, operation):
        """
        Handle remove operations.
        """
        if path == ('members', None, None):
            members = value or []
            ids = [int(member.get('value')) for member in members]
            users = get_user_model().objects.filter(id__in=ids)

            if len(ids) != users.count():
                raise exceptions.BadRequestError('Can not remove a non-existent user from group')

            for user in users:
                self.obj.user_set.remove(user)

        else:
            raise exceptions.NotImplementedError

    def handle_replace(self, path, value, operation):
        """
        Handle the replace operations.
        """
        if path == ('name', None, None):
            name = value[0].get('value')
            self.obj.name = name
            self.obj.save()

        else:
            raise exceptions.NotImplementedError

